<?php
namespace Ekliptor\Wordpress;

class DatabaseMigration {
	/** @var string Latest DB version. Only updated with plugin version if there are migrations. */
	protected $lastVersion;
	/** @var string */
	protected $currentVersion;
	/** @var array */
	protected $lastError = array();
	
	public function __construct(string $lastVersion, string $currentVersion) {
		if (!$lastVersion)
			$lastVersion = '1.2.22'; // when we added migration. shouldn't be needed
		$this->lastVersion = $lastVersion;
		$this->currentVersion = $currentVersion;
	}
	
	public static function checkAndMigrate() {
		$lastVersion = get_option('slp_payments_installed');
		if ($lastVersion === BCHSLP_PAYMENTS_VERSION)
			return;
		add_action('plugins_loaded', function() use ($lastVersion) {
			$migrate = new DatabaseMigration($lastVersion, BCHSLP_PAYMENTS_VERSION);
			try {
				if ($migrate->migrate() === false) {
					SlpPaymentsCore::notifyErrorExt("Error ensuring latest DB version on migration", $migrate->getLastError());
					return;
				}
				update_option( 'slp_payments_installed', BCHSLP_PAYMENTS_VERSION ); // should already be done in main plugin class
			}
			catch (\Exception $e) {
				SlpPaymentsCore::notifyErrorExt("Exception during DB migration: " . get_class(), $e->getMessage());
			}
		}, 200); // load after other plugins
	}
	
	public function migrate(): bool {
		$queries = array();
		// TODO also add crons here if we add more later
		switch ($this->lastVersion) {
			// add migration queries in order from oldest version to newest
			case '1.2.22':
				//$table = SlpPaymentsCore::getTableName('transactions');
				//if ($this->columnExists($table, 'address') === false) {
					//$queries[] = "ALTER TABLE `$table` ADD `amount` FLOAT NOT NULL AFTER `address`";
				//}
				add_action('shutdown', function() {
					$gateway = SlpPayments::getInstance(); // setting class has its own migrations, but just to be sure for this important one
					$gateway->getSettings()->set('slpSocketAddress', 'https://slpsocket.fountainhead.cash/s/');
					$gateway->settings['slpSocketAddress'] = $gateway->getSettings()->get('slpSocketAddress');
				});
		}
		if (empty($queries))
			return true; // say successful
		return $this->runQueries($queries);
	}
	
	public function getLastError(): array {
		return $this->lastError;
	}
	
	protected function runQueries(array $queries): bool {
		global $wpdb;
		foreach ($queries as $query) {
			$result = $wpdb->query($query);
			if ($result === false) {
				$this->lastError = array(
						'query' => $query,
						'error' => $wpdb->last_error
				);
				return false; // abort
			}
		}
		return true;
	}
	
	protected function columnExists(string $table, string $column) {
		global $wpdb;
		$rows = $wpdb->query("SHOW COLUMNS FROM `$table` LIKE '$column'");
		return empty($rows) ? false : true;
	}
}
?>