<?php
namespace Ekliptor\Wordpress;

use Ekliptor\CashP\BlockchainApi\AbstractBlockchainApi;
use Ekliptor\CashP\BlockchainApi\Http\WordpressHttpAgent;

include BCHSLP_PAYMENTS__PLUGIN_DIR . 'classes/SlpPaymentsCore.php';
include BCHSLP_PAYMENTS__PLUGIN_DIR . 'classes/SlpPaymentsApi.php';


/**
 * The Woocommerce gateway implementation.
 * This class servers as the entry point and the constructor gets called from Woocommerce.
 * It creates a SlpPaymentsCore instance for all functionality that can be used beyong the scope of WC in WP.
 *
 */
class SlpPayments extends \WC_Payment_Gateway {
	const WC_ADMIN_PAGE = 'admin.php?page=wc-settings&tab=checkout&section=slp_payments_woocommerce';
	const CHECK_CONFIRMATIONS_WEB_MIN = 10;
	const CHECK_PAYMENT_INTERVAL_SEC = 30;
	const SHOW_ORDER_EXPIRATION_MAX_H = 24; // show payment bust be made within that timeframe if WC manage stock setting is below
	const ORDER_UNPAID_SKIP_CHECK_SEC = 3;
	
	/** @var SlpPayments */
	private static $instance = null;
	private static $cron_events = array (
		);
	private static $cron_events_hourly = array(
			'slp_check_transaction_confirmations'
		);
	
	/** @var SlpPaymentsCore */
	protected $slpPaymentsCore;
	/** @var Settings */
	protected $pluginSettings; // settings already used in parent class
	/** @var TemplateEngine */
	public $tpl = null;
	
	/** @var AbstractBlockchainApi */
	protected $blockchainApi = null;
	
	/** SLPDB as a faster alternative to Bitcoin.com REST DB.
	 *  @var AbstractBlockchainApi */
	protected $slpDbApi = null;
	
	/** @var \WC_Session|\WC_Session_Handler */
    protected $session = null;
    /** @var \WC_Cart */
    //protected $cart = null;
    /** @var bool */
    protected $paymentOptionsShowing = false;
    /** @var bool */
    protected $showDebugInfo = false;
	
	public function __construct(/*\SlpPaymentsCore $slpPaymentsCore*/) {
		// this gets called from Woocommerce, so make sure we cache this instance
		static::check_plugin_activation();
		//$this->slpPaymentsCore = $slpPaymentsCore;
		$this->slpPaymentsCore = SlpPaymentsCore::getInstance();
		if (self::$instance === null)
			self::$instance = $this;
		$this->showDebugInfo = WP_DEBUG || SlpPaymentsCore::DEBUG;
		
		$this->id            		= 'slp_payments_woocommerce';
        $this->medthod_title 		= __('SLP Payments for WooCommerce', 'ekliptor');
        $this->has_fields    		= true;
        $this->method_description 	= __('Accept SLP tokens and Bitcoin Cash in your online store as payment.', 'ekliptor');
        $this->has_fields			= true;
        
        $this->init();
        $this->slpPaymentsCore->init(); // must be done after constructor
        
        // use the first token image if we have an enabled token. good idea?
        // better always use the same SLP image for the gateway
        //$this->icon					= !empty($this->settings['tokenImage']) ? $this->settings['tokenImage'] : plugins_url( 'img/slp_48.png', BCHSLP_PAYMENTS__PLUGIN_DIR . 'slp-payments.php' );
        $this->icon					= plugins_url( 'img/SLP-logo-stroke-48.png', BCHSLP_PAYMENTS__PLUGIN_DIR . 'slp-payments.php' );
        $title = isset($this->settings['title']) ? $this->settings['title'] : $this->getFrontendDefaultTitle();
        $description = isset($this->settings['description']) ? $this->settings['description'] : $this->getFrontendDefaultDescription();
        $this->title       			= $title; // for frontend (user), also shown in WC settings
        $this->description 			= $description; // for frontend (user)
        //$this->order_button_text 	= "fooo..."; // TODO add option to replace the whole button HTML by overwriting parent functions
        
        $this->session 				= WC()->session;
        //$this->cart    				= WC()->cart;
        
        //pre_print_r($this->blockchainApi->createNewAddress($this->pluginSettings->get('xPub'), 0));
        //pre_print_r($this->blockchainApi->getAddressDetails('bitcoincash:qzu6azekp8q49ydrt5hgf3wnhqjtk9hkjv5q3fzwnu'));
        //pre_print_r($this->blockchainApi->getSlpAddressDetails('simpleledger:qzu6azekp8q49ydrt5hgf3wnhqjtk9hkjvcm6jhwdz', '7278363093d3b899e0e1286ff681bf50d7ddc3c2a68565df743d0efc54c0e7fd'));
        //$this->sendOrderReceivedEmail(951);
	}
	
	public static function getInstance(/*\SlpPaymentsCore $slpPaymentsCore = null*/) {
		if (self::$instance === null)
			self::$instance = new self(/*$slpPaymentsCore*/);
		return self::$instance;
	}
	
	public function init() {
		$this->pluginSettings = $this->slpPaymentsCore->getSettings();
		//$this->pluginSettings->set('blockchain_api', 'BitcoinComRestApi');
		$this->blockchainApi = $this->slpPaymentsCore->createBlockchainApiInstance();
		if (is_admin() === true) {
			$this->tpl = new TemplateEngine($this->pluginSettings);
		}
		
		//$this->set_defaults(array('slpSocketAddress' => '')); // already done on field init
		$this->init_settings();
		$this->init_form_fields();
		
		// init hooks
		// note that functions must be public because they are called from event stack
		// call the main class which will call this plugin via our own action
		
		add_filter('slp_payments_js_config', array($this, 'addJavascriptConfig'), 10, 1);
		add_filter('slp_default_settings', array($this, 'addDefaultSettings'), 10, 1);
		
		// WC gateway hooks
		//add_action('woocommerce_api_wc_coinpay', array($this, 'checkIpnResponse')); // called after payment if we use a 3rd party callback
		add_action('woocommerce_update_options_payment_gateways_' . $this->id, array($this, 'process_admin_options'));
		
		// different WC plugins might overwrite the order details page. so register for all and only print data once with the first hook (far up on the page)
		add_action('woocommerce_order_details_before_order_table', array ($this, 'addPluginPaymentOptions' ), 100, 1);
		add_action('woocommerce_order_details_after_order_table', array ($this, 'addPluginPaymentOptions' ), 100, 1);
		add_action('woocommerce_order_details_before_order_table_items', array ($this, 'addPluginPaymentOptions' ), 100, 1);
		//add_action('woocommerce_thankyou', array ($this, 'addPluginPaymentOptions' ), 100, 1);
		add_action('woocommerce_thankyou_' . $this->id, array ($this, 'addPluginPaymentOptions' ), 100, 1);
		
		add_filter('woocommerce_settings_api_sanitized_fields_' . $this->id, array($this, 'onSettingsUpdate'));
		
		// WC email hooks
		add_action( 'woocommerce_email_before_order_table', array($this, 'addBlockchainDataToEmail'), 10, 4 );
		add_action( 'woocommerce_order_status_pending', array($this, 'sendOrderReceivedEmail'), 10, 1 ); // unreliable as this plugin might not be instantiated when this hook triggers
		
		// admin messages
		if (is_admin() === true) {
			// plugin gets registered after admin_init action
			$this->displayAdminNotices();
			$this->loadAssets();
			add_action( 'woocommerce_display_admin_footer_text', array($this, 'addWoocommerceAdminFooterCode') );
		}
		
		// Crons
		//add_action ( 'slp_check_transaction_confirmations', array ($this, 'checkTransactionConfirmations' ) );
		
		$lastCheck = $this->pluginSettings->get('lastCheckedTransactions');
		if ($lastCheck + static::CHECK_CONFIRMATIONS_WEB_MIN*60 <= time()) {
        	//$this->checkTransactionConfirmations(); // better do it at the end of script (to be shown on next page load)
        	add_action('shutdown', array ($this, 'checkTransactionConfirmations' ));
		}
	}
	
	public static function check_plugin_activation() {
		$pluginVersion = get_option ( 'slp_payments_installed' );
		if ($pluginVersion === BCHSLP_PAYMENTS_VERSION)
			return;
		SlpPaymentsCore::plugin_activation();
		
		foreach ( self::$cron_events as $cron_event ) {
			$timestamp = wp_next_scheduled ( $cron_event );
			if (!$timestamp)
				wp_schedule_event(time(), 'daily', $cron_event);
		}
		foreach ( self::$cron_events_hourly as $cron_event ) {
			$timestamp = wp_next_scheduled ( $cron_event );
			if (!$timestamp)
				wp_schedule_event(time(), 'hourly', $cron_event);
		}
		
		update_option ( 'slp_payments_installed', BCHSLP_PAYMENTS_VERSION );
	}
	
	public static function plugin_deactivation() {
		// Remove any scheduled cron jobs.
		$events = array_merge(self::$cron_events, self::$cron_events_hourly);
		foreach ( $events as $cron_event ) {
			$timestamp = wp_next_scheduled ( $cron_event );
			if ($timestamp) {
				wp_unschedule_event ( $timestamp, $cron_event );
			}
		}
		
		SlpPaymentsCore::plugin_deactivation();
		delete_option('slp_payments_installed');
	}
	
	public static function addWoocommerceGateway(array $load_gateways) {
		$load_gateways[] = '\\Ekliptor\\Wordpress\\SlpPayments';
		return $load_gateways;
	}
	
	public static function setupCronHooks() {
		SlpPaymentsCore::setupCronHooks();
		add_action ( 'slp_check_transaction_confirmations', array (static::getInstance(), 'checkTransactionConfirmations' ) );
	}
	
	public function getCore(): SlpPaymentsCore {
		return $this->slpPaymentsCore;
	}
	
	public function getSettings(): Settings {
		return $this->pluginSettings;
	}
	
	public function init_settings() {
		parent::init_settings();
		$this->enabled  = ! empty( $this->settings['enabled'] ) && 'yes' === $this->settings['enabled'] ? 'yes' : 'no';
		if (!isset($this->settings['tokenID']) || empty($this->settings['tokenID'])) { // set the first token on 1st admin page open
			$preInstalledTokens = PreInstalledTokens::getWoocommerceSettingsArray();
			$tokenSymbols = array_keys($preInstalledTokens);
			$firstToken = $preInstalledTokens[$tokenSymbols[0]];
			foreach ($firstToken as $key => $value) {
				$this->settings[$key] = $value;
			}
		}
		if (empty($this->settings['xPub']))
			$this->settings['xPub'] = $this->slpPaymentsCore->getSettings()->get('xPub');
	}
	
	public function init_form_fields() {
		$this->form_fields = $this->getFormFieldArray();
		// to use custom form controls (such as a table) we could overwrite generate_settings_html() and validate input manually
    }
    
    public function is_available() {
    	if (empty($this->slpPaymentsCore->getSettings()->get('xPub')))
    		return false;
    	if (/*isset($this->settings['enabledToken']) && $this->settings['enabledToken'] === 'yes' && */empty($this->settings['tokenID']))
    		return false;
    	return parent::is_available();
    }
    
    public function validate_text_field(/*string */$key, $value) { // validate_description_field
    	$value = trim($value);
    	if (!empty($value))
    		return $value;
    	$fields = $this->getFormFieldArray();
    	if (isset($fields[$key]) && isset($fields[$key]['default'])) {
    		$field = $fields[$key];
    		$useFieldType = is_string($field['default']) || is_int($field['default']) || is_float($field['default']);
    		if ($useFieldType === true)
    			return $field['default'];
    	}
    	if ($key === 'description')
    		return $this->getFrontendDefaultDescription();
    	else if ($key === 'title')
    		return $this->getFrontendDefaultTitle();
    	return '';
    }
    
    public function payment_fields() {
    	// TODO set input type="number" and step size based on SLP decimals
    	// TODO support more currencies than USD. for this we also need a backend like bitcoin.com to crawl BCH prices
    	$tokens = $this->slpPaymentsCore->getEnabledTokens();
    	if (/*$this->settings['enabledToken'] !== 'yes'*/empty($tokens)) {
    		echo esc_html($this->description);
    		return;
    	}
    	// TODO update html + order object to allow multiple payments
    	$formVars = array(
    			'description' => $this->description,
    			//'tokenName' => $this->pluginSettings->get('tokenName'),
    			//'tokenTickerSymbol' => $this->pluginSettings->get('tokenTickerSymbol'),
    			//'tokenRate' => $this->pluginSettings->get('token_rate_usd'),
    			//'tokenImage' => $this->pluginSettings->get('tokenImage'),
    			//'tokenWebsite' => $this->pluginSettings->get('tokenWebsite'),
    			'tokens' => $tokens,
    			'bchRate' => $this->pluginSettings->get('rate_usd_bch'),
    			'baseCurrency' => $this->pluginSettings->get('button_currency'),
    			'allowBchPayments' => $this->settings['allowBchPayments'] === 'yes',
    	);
    	include BCHSLP_PAYMENTS__PLUGIN_DIR . 'tpl/client/payForm.php';
    }
    
    /**
     * The "place order" has just been clicked. This is the moment the order has been created and we can instantiate an order object by ID
     * {@inheritDoc}
     * @see \WC_Payment_Gateway::process_payment()
     */
    public function process_payment($order_id) {
    	/*
    	if (!isset($_POST['slp-token-amount'])) {
    		wc_add_notice(esc_html("Missing payment form parameters.", "ekliptor"), 'error');
    		return;
    	}
    	else */if (!$this->session) { // shouldn't happen (and if it does the cart will be empty too)
    		wc_add_notice(esc_html("Your session has expired. You have not been charged. Please add your item(s) to the cart again.", "ekliptor"), 'error');
    		return;
    	}
    	//$bchAddress = $this->session->get("bchAddress")->cashAddress;
    	$this->clearCustomSessionVariables(); // ensure there is no plugin data left from the previous payment
    	$this->session->set("orderID", $order_id);
    	$order = new \WC_Order($order_id);
    	
    	// decide how many tokens and/or BCH shall be paid
    	$totalFiat = $order->get_total();
    	//$tokenAmount = isset($_POST['slp-token-amount']) && $_POST['slp-token-amount'] > 0.0 ? (float)$_POST['slp-token-amount'] : 0.0;
    	$tokenAmount = 0.0; // we don't let the user enter amount to split payments currently
    	$tokenSymbol = '';
    	if (isset($_POST['slp-payment-method']) && $tokenAmount = $_POST['slp-payment-method'] !== 'slp-bch')
    		$tokenSymbol = strtoupper(substr($_POST['slp-payment-method'], 4));
    	// new method: radio button for all in tokens or BCH
    	//if ($this->settings['enabledToken'] === 'yes') {
    		if (isset($_POST['slp-payment-method']) && $tokenAmount = $_POST['slp-payment-method'] !== 'slp-bch')
    			$tokenAmount = $this->slpPaymentsCore->getTokenAmount($totalFiat, $tokenSymbol);
    		else if ($this->settings['allowBchPayments'] !== 'yes') { // BCH payments disabled, we must use a token
    			$tokenAmount = $this->slpPaymentsCore->getTokenAmount($totalFiat, $tokenSymbol); // prevent setting overwrite via forged POST
    			$enabledTokens = $this->slpPaymentsCore->getEnabledTokens();
    			$tokenSymbols = array_keys($enabledTokens);
				$tokenSymbol = $enabledTokens[$tokenSymbols[0]]; // use the first enabled token
    		}
    	//}
    	
    	$slpToken = $tokenAmount > 0.0 ? $this->slpPaymentsCore->getTokenBySymbol($tokenSymbol) : null;
    	$bchAmount = $this->slpPaymentsCore->getRemainingBchAmount($tokenAmount, $totalFiat, $slpToken ? $slpToken['symbol'] : "");
    	//$this->session->set("tokenAmount", $tokenAmount);
    	$order->add_meta_data('bchAmount', $bchAmount, true);
    	$order->add_meta_data('tokenAmount', $tokenAmount, true);
    	if ($slpToken) {
    		$order->add_meta_data('tokenID', $slpToken['id'], true);
    		$order->add_meta_data('tokenSymbol', $tokenSymbol, true);
    	}
    	$order->add_meta_data('tokenAmountReceived', 0.0, true);
    	$order->add_meta_data('bchAmountReceived', 0.0, true);
    	$order->save_meta_data();
    	
    	/*
    	if ($this->isPaidOrder($order_id, false) === true) {
    		sleep(2);
    		if ($this->isPaidOrder($order_id, true) === true) // retry to avoid concurrency issues
    			return;
    	}
    	*/
    	//$wcOrder = wc_get_order($order_id);
    	//$this->sendOrderReceivedEmail($order_id); // better do it after redirect on payment page
    	
		return array(
				'result' => 'success',
				'redirect' => $this->get_return_url($order) // just redirect to the order details page, whe show our payment button(s) there
		);
	}
	
	public function onSettingsUpdate(array $settings) {
		$updateSettings = array(); // update settings in a single DB query
		foreach ($settings as $key => $value) { // remove temp display settings
			if (substr($key, 4) === '_dbg')
				unset($settings[$key]);
		}
		if (isset($settings['xPub'])) {
			if ($this->settings['xPub'] !== $settings['xPub']) {
				// reset the hdPath counter (otherwise electron cash needs to scan the full history for the TX to show up
				//$this->pluginSettings->set('addressCount', 0);
				$updateSettings['addressCount'] = 0;
			}
			$updateSettings['xPub'] = $settings['xPub'];
			$this->slpPaymentsCore->scheduleUnsuedAddressSearch(); // receiving xPub changed. search for unused address range start
		}
		$removeButtonSettings = array('addToken', 'deleteToken');
		foreach ($removeButtonSettings as $key) {
			if (isset($settings[$key]))
				unset($settings[$key]);
		}
		$allTokenJson = json_decode(stripcslashes($settings['allTokenJson']));
		unset($settings['allTokenJson']);
		
		if (!isset($settings['tokenID']))
			$settings['tokenID'] = ''; // shouldn't happen (unless admin edits html)
		if (isset($settings['blockchainApi'])) {
			if (SlpPaymentsCore::isValidUrl($settings['blockchainApi']) === false) // restore previous if invalid
				$settings['blockchainApi'] = $this->pluginSettings->get('blockchain_rest_url');
			else
				$updateSettings['blockchain_rest_url'] = $settings['blockchainApi'];
		}
		if (isset($settings['slpDbApiUrl'])) {
			if (SlpPaymentsCore::isValidUrl($settings['slpDbApiUrl']) === false) // restore previous if invalid
				$settings['blockchainApi'] = $this->pluginSettings->get('slpDbApiUrl');
			else
				$updateSettings['slpDbApiUrl'] = $settings['blockchainApi'];
		}
		if (isset($settings['slpSocketAddress'])) {
			if (SlpPaymentsCore::isValidUrl($settings['slpSocketAddress']) === false) // restore previous if invalid
				$settings['slpSocketAddress'] = $this->pluginSettings->get('slpSocketAddress');
			else
				$updateSettings['slpSocketAddress'] = $settings['slpSocketAddress'];
		}
		
		$settings['tokenTickerSymbol'] = strtoupper($settings['tokenTickerSymbol']);
		// TODO add more validaion such as not allow "BCH" as a token symbol
		if (isset($settings['tokenImage']) && SlpPaymentsCore::isValidUrl($settings['tokenImage']) === false) {
			$settings['tokenImage'] = '';
		}
		
		// add a new token?
		$symbol = $settings['tokenTickerSymbol'];
		$preInstalledTokens = PreInstalledTokens::getTokens();
		$availableTokens = $this->pluginSettings->get('availableTokens');
		if (isset($settings['addingNewToken']) && $settings['addingNewToken'] === '1') {
			unset($settings['addingNewToken']); // no need to store it
			$settings['tokenSelect'] = $settings['tokenTickerSymbol']; // select the new token
			//$updateSettings['reloadSettings'] = true;
		}
		
		// ensure pre-installed tokens are always present and shown on top
		$availableTokensUpdated = array();
		$tokenDisabled = !isset($settings['enableCurrentToken']) || $settings['enableCurrentToken'] !== 'yes' ? true : false;
		foreach ($preInstalledTokens as $curSymbol => $preToken) {
			//if (isset($availableTokens[$curSymbol])) // safer to just use always the hardcoded value
				//continue;
			$availableTokensUpdated[$curSymbol] = array(
					'id' => $preToken->id,
					'name' => $preToken->name,
					'symbol' => $preToken->symbol,
					'documentUri' => $preToken->documentUri,
					'icon' => $preToken->icon,
					'rate' => $preToken->rate,
					'enabled' => $preToken->enabled,
			);
			if ($symbol === $curSymbol) { // WP (HTML forms) doesn't submit disabled input controls here, so we have to fill in pre-installed values
				//$wcTokens = PreInstalledTokens::getWoocommerceSettingsArray(); // same properties as here
				$settings['tokenID'] = $preToken->id;
				$settings['tokenName'] = $preToken->name;
				$settings['tokenTickerSymbol'] = $preToken->symbol; // TODO not needed anymore
				$settings['tokenWebsite'] = $preToken->documentUri;
				$settings['tokenImage'] = $preToken->icon;
				$settings['tokenRateUSD'] = $preToken->rate;
				$settings['enableCurrentToken'] = $tokenDisabled ? 'no' : 'yes';
			}
		}
		
		// add user tokens
		$availableTokens = array_merge($availableTokensUpdated, (array)$allTokenJson, $availableTokens);
		foreach ($availableTokens as $curSymbol => $curToken) {
			if (is_numeric($curToken['rate']) === false || $curToken['rate'] < 0.0)
				$curToken['rate'] = 1.0; // ensure positive rates
		}
		
		// delete the current token?
		$deleteToken = isset($settings['deleteCurrentToken']) && $settings['deleteCurrentToken'] === '1';
		if ($deleteToken === true) {
			unset($settings['deleteCurrentToken']); // no need to store it
			unset($availableTokens[$symbol]);
			$settings['tokenSelect'] = array_keys($preInstalledTokens)[0]; // select the first token
			//$updateSettings['reloadSettings'] = true;
		}
		if (isset($availableTokens['']))
			unset($availableTokens['']);
		
		// add or update this token
		if (/*!isset($availableTokens[$symbol]) && */!isset($preInstalledTokens[$symbol]) && $deleteToken === false) { // add it if it's a user-defined token
			$availableTokens[$symbol] = array(
					'id' => $settings['tokenID'],
					'name' => $settings['tokenName'],
					'symbol' => $settings['tokenTickerSymbol'],
					'documentUri' => $settings['tokenWebsite'],
					'icon' => $settings['tokenImage'],
					'rate' => $settings['tokenRateUSD'],
					'enabled' => $tokenDisabled ? false : true
			);
		}
		// user can't remove pre-installed tokens, but disable them for customers
		if ($deleteToken === false)
			$availableTokens[$symbol]['enabled'] = $tokenDisabled ? false : true;
		
		// removed root props and only use sub props everyhwere
		if (isset($settings['tokenRateUSD'])) {
			//$updateSettings['token_rate_usd'] = $settings['tokenRateUSD'];
			$availableTokens[$symbol]['rate'] = $settings['tokenRateUSD'];
		}
		if (isset($settings['tokenImage'])) {
			//$updateSettings['icon'] = $settings['tokenImage'];
			$availableTokens[$symbol]['icon'] = $settings['tokenImage'];
		}
		$copyPluginSettings = array('tokenID', 'tokenName', 'tokenTickerSymbol', 'tokenWebsite');
		foreach ($copyPluginSettings as $key) {
			$availableTokens[$symbol][$key] = $settings[$key];
		}
		
		// crawl the token name & symbol if not provided by admin. always crawl it because we need the number of decimal places
		//$this->pluginSettings->set('enabledToken', isset($settings['enabledToken']) && $settings['enabledToken'] === 'yes' && isset($settings['tokenID']) && !empty($settings['tokenID']));
		if (!empty($settings['tokenID'])/* && (empty($settings['tokenTickerSymbol']) || empty($settings['tokenName']))*/) {
			$token = $this->blockchainApi->getTokenInfo($settings['tokenID']);
			if ($token !== null) {
				if (empty($settings['tokenTickerSymbol']))
					$settings['tokenTickerSymbol'] = $token->symbol;
				if (empty($settings['tokenName']))
					$settings['tokenName'] = $token->name;
				$availableTokens[$symbol]['decimals'] = $token->decimals >= 0 ? $token->decimals : 8;
			}
		}
		
		$updateSettings['availableTokens'] = $availableTokens;
		$this->pluginSettings->setMultiple($updateSettings);
		return $settings;
	}
	
	public function displayAdminNotices() {
		if (!isset($this->settings['enabled']) || $this->settings['enabled'] !== 'yes')
			return;
		if (empty($this->settings['xPub'])) {
			$tplVars = array(
					'msg' => __('You must enter the xPub of your BCH wallet to enable SLP payments.', 'ekliptor'),
					'link' => admin_url() . static::WC_ADMIN_PAGE
			);
			$notice = new AdminNotice($this->tpl->getTemplate('adminNotice.php', $tplVars), 'error');
			$this->tpl->addAdminNotices($notice);
		}
		if (/*$this->settings['enabledToken'] === 'yes' && */empty($this->settings['tokenID'])) {
			$tplVars = array(
					'msg' => __('You must enter the token ID of your SLP token to receive payments.', 'ekliptor'),
					'link' => admin_url() . static::WC_ADMIN_PAGE
			);
			$notice = new AdminNotice($this->tpl->getTemplate('adminNotice.php', $tplVars), 'error');
			$this->tpl->addAdminNotices($notice);
		}
		/*
		if (isset($_GET['notices'])) {
			$notices = explode(',', $_GET['notices']);
			foreach ($notices as $noticeData) {
				$notice = AdminNotice::urlDecode($noticeData);
				$this->tpl->addAdminNotices($notice);
			}
		}
		do_action('slp_payments_admin_notices');
		*/
		add_action('admin_notices', array($this->tpl, 'showAdminNotices'));
	}
	
	public function loadAssets() {
		//wp_enqueue_style( 'slp-payments-admin', plugins_url( 'tpl/css/slp-payments-admin.css', BCHSLP_PAYMENTS__PLUGIN_DIR . 'slp-payments.php' ), array(), BCHSLP_PAYMENTS_VERSION );
		wp_enqueue_script( 'slp-payments-bundle', plugins_url( 'tpl/js/bundle.js', BCHSLP_PAYMENTS__PLUGIN_DIR . 'slp-payments.php' ), array(), BCHSLP_PAYMENTS_VERSION, false );
	}
	
	public function addPluginPaymentOptions($order_id) {
		if ($this->paymentOptionsShowing === true)
			return;
		if (!$this->session)
			return; // shouldn't happen
		if (!$order_id) {
			if (!$this->session->get("orderID"))
				return;
			$order_id = $this->session->get("orderID");
		}
		$this->paymentOptionsShowing = true;
		try {
			$order = is_object($order_id) ? $order_id : new \WC_Order($order_id); // some hooks return the order object already
			$this->session->set("orderID", $order->get_id()); // ensure it's the current order
		}
		catch (\Exception $e) { // invalid order exception
			echo esc_html('This order does not exist.', 'ekliptor') . '<br><br>';
			return;
		}
		
		if ($order->get_payment_method() !== 'slp_payments_woocommerce')
			return; // the user chose another payment method
		$this->isPaidOrder($order); // trigger the check on manual page reload
		if ($order->get_status() === 'cancelled')
			return;
		if ($order->is_paid() === true) {
			echo esc_html('Your order has been fully paid.', 'ekliptor') . '<br><br>';
			return;
		}
		
		// call isPaidOrder() here again (for older orders) and update status if user manually reloads this page to get the payment completed before cron?
		$this->addChainTransactionID($order);
		$bchTxID = $order->get_meta('bchTransactionID');
		//if ($bchTxID) { // user says the payment has already been sent
			if ($this->checkPaidOrder($order) === false) {
				if ($bchTxID) {
					if ($this->settings['enable0Conf'] === 'yes') // we accepted the payment already and set the order status to "processing"
						esc_html_e('Please wait while your payment is being broadcasted to the network. You may close this page or reload it in a few seconds for updates.', 'ekliptor');
					else
						printf(esc_html('Your payment is waiting for %s confirmations. You may close this page and check the store again in couple of minutes. Alternatively just reload this page.', 'ekliptor'), $this->pluginSettings->get('wait_confirmations'));
					echo "<br><br>";
				}
			}
			else {
				$order->update_status($this->getSuccessPaymentStatus());
				$order->save_meta_data();
				return;
			}
		//}
		
		$btnConf = array();
		$includedButtonScript = $this->slpPaymentsCore->getIncludedButtonScript();
		$btnConf['recAddress'] = $this->session->get("bchAddress");
		//$btnConf['useTokenPayments'] = $this->slpPaymentsCore->useTokenPayments();
		$btnConf['useTokenPayments'] = (float)$order->get_meta('tokenAmount') > 0.0;
		$btnConf['tokenID'] = $order->get_meta('tokenID');
		if (empty($btnConf['recAddress'])) {
			$nextCount = $this->pluginSettings->get('addressCount') + 1;
			$this->pluginSettings->set('addressCount', $nextCount);
			$tokenID = $btnConf['useTokenPayments'] === true ? $btnConf['tokenID'] : '';
			$btnConf['recAddress'] = $this->blockchainApi->createNewEmptyAddress($this->settings['xPub'], $nextCount, $this->pluginSettings->get('hdPathFormat'), $tokenID);
			if ($btnConf['recAddress'] === null) {
				echo esc_html("Unable to create a new address. Please try again or report a bug if the problem persists.", "ekliptor") . '<br><br>';
				return;
			}
			// store BCH address in session as it is unique for this customer
			$this->pluginSettings->set('lastAddress', $btnConf['recAddress']->cashAddress);
			$this->session->set("bchAddress", $btnConf['recAddress']); // TODO also store in DB to have same address if user deletes session (clears cookies)
		}
		$btnConf['recAddress'] = $btnConf['useTokenPayments'] === true ? $btnConf['recAddress']->slpAddress : $btnConf['recAddress']->cashAddress;
		if (!$order->get_meta('chainAddress')) {
			$chainAddrress = $this->session->get("bchAddress");
			$order->add_meta_data('slpPayment', $btnConf['useTokenPayments'] === true ? '1' : '0', true);  // TODO move to a better place?
			$order->add_meta_data('chainAddress', $chainAddrress->cashAddress, true); // always BCH address. we need both for checking balances
			$order->add_meta_data('slpAddress', $chainAddrress->slpAddress, true);
			$order->save_meta_data();
		}
		
		$btnConf['unit'] = $this->pluginSettings->get('button_currency');
		$btnConf['amount'] = $order->get_total();
		if ($btnConf['amount'] < 0.00000001)
			$btnConf['amount'] = 0.00000001;
		$btnConf['sats'] = SlpPaymentsCore::toSatoshis($btnConf['amount'] / $this->pluginSettings->get('rate_usd_bch'));
		//$btnConf['amountBCH'] = $this->slpPaymentsCore->toAmountBCH($btnConf['amount'], $btnConf['unit']);
		$btnConf['amountBCH'] = (float)$order->get_meta('bchAmount');
		if (!$btnConf['amountBCH'])
    		$btnConf['amountBCH'] = 0.0;
		//$btnConf['txid'] = $this->session->get("txid");
		$btnConf['txid'] = $order->get_meta('slp-txid');
		
		$slpToken = $btnConf['useTokenPayments'] === true ? $this->slpPaymentsCore->getTokenBySymbol($order->get_meta('tokenSymbol')) : null;
		
		if (empty($btnConf['txid'])) {
			if ($slpToken)
				$btnConf['txid'] = $this->slpPaymentsCore->createTransactionId($btnConf['recAddress'], $btnConf['amount'], (float)$order->get_meta('tokenAmount'), $order->get_id(), $slpToken['id'], $slpToken['symbol']);
			else
				$btnConf['txid'] = $this->slpPaymentsCore->createTransactionId($btnConf['recAddress'], $btnConf['amount'], (float)$order->get_meta('bchAmount'), $order->get_id());
			if ($btnConf['txid'] === false) {
				echo esc_html("Unable to create a transaction ID. Please try again or report a bug if the problem persists.", "ekliptor") . '<br><br>';
				return;
			}
			//$this->session->set("txid", $btnConf['txid']);
			$order->add_meta_data('slp-txid', $btnConf['txid'], true);
			$order->save_meta_data();
		}
		
		$btnConf['slpIcon'] = plugins_url( 'img/slp_48.png', BCHSLP_PAYMENTS__PLUGIN_DIR . 'slp-payments.php' );
		$btnConf['addQrCode'] = true;
		//$btnConf['qrcodeStatic'] = ''; // we show the real code right away
		$btnConf['qrcode'] = $this->slpPaymentsCore->generateQrCode($btnConf['txid'], $btnConf['amountBCH']);
		$btnConf['btnText'] = __('Pay with Badger Wallet', 'ekliptor');
		if ($btnConf['useTokenPayments'] === true) {
			$btnConf['tokenName'] = $slpToken['name'];
    		$btnConf['tokenTickerSymbol'] = $order->get_meta('tokenSymbol');
    		$btnConf['tokenImage'] = $slpToken['icon'];
    		$btnConf['tokenWebsite'] = $slpToken['documentUri'];
    		//$btnConf['tokenAmount'] = $this->session->get("tokenAmount");
    		$btnConf['tokenAmount'] = (float)$order->get_meta('tokenAmount');
		}
		else {
			$btnConf['tokenName'] = '';
    		$btnConf['tokenTickerSymbol'] = 'BCH';
    		$btnConf['tokenImage'] = '';
    		$btnConf['tokenWebsite'] = '';
    		$btnConf['tokenAmount'] = '';
		}
    	if (!$btnConf['tokenAmount'])
    		$btnConf['tokenAmount'] = 0.0;
    	if ($slpToken)
    		$btnConf['uri'] = $this->slpPaymentsCore->createPaymentURI($btnConf['recAddress'], $btnConf['amountBCH'], $btnConf['tokenAmount'], $slpToken['id'], isset($slpToken['decimals']) ? $slpToken['decimals'] : 8);
    	else
    		$btnConf['uri'] = $this->slpPaymentsCore->createPaymentURI($btnConf['recAddress'], $btnConf['amountBCH'], 0.0);
		$btnConf['callbackData'] = array(
				'slpPayments' => true,
				'woocommerce' => true
		);
		$btnConf['addBadger'] = $this->settings['enableBadger'] === 'yes';
		$btnConf['loadingImage'] = plugins_url( 'img/loading2.gif', BCHSLP_PAYMENTS__PLUGIN_DIR . 'slp-payments.php' );
		$btnConf['copyImage'] = plugins_url( 'img/copy.svg', BCHSLP_PAYMENTS__PLUGIN_DIR . 'slp-payments.php' );
		
		// TODO move this inline-css class to new styles.css.php file if we have more dynamic css
		$btnConf['preloaderImage'] = plugins_url( 'img/icon-preloader-dots2.svg', BCHSLP_PAYMENTS__PLUGIN_DIR . 'slp-payments.php' );
		
		$wc_held_duration = floor(get_option( 'woocommerce_hold_stock_minutes' ) / 60);
		$btnConf['orderTimeoutTxt'] = '';
		if ( $wc_held_duration >= 1 && $wc_held_duration <= static::SHOW_ORDER_EXPIRATION_MAX_H && 'yes' === get_option( 'woocommerce_manage_stock' ) ) // be safer and only check for the time setting as admin might change that after order?
			$btnConf['orderTimeoutTxt'] = ' ' . sprintf(__('You must pay within %s hours.', 'ekliptor'), $wc_held_duration);
		
		// headline text
		if ($btnConf['useTokenPayments'] === true) {
			$btnConf['amountDisplay'] = number_format($btnConf['tokenAmount'], $slpToken['decimals']);
		}
		else {
			$btnConf['amountDisplay'] = number_format($btnConf['amountBCH'], $this->pluginSettings->get('display_digits_bch'));
		}
		$btnConf['tickerDisplay'] = $btnConf['useTokenPayments'] === true ? $btnConf['tokenTickerSymbol'] : 'BCH';
		
		include BCHSLP_PAYMENTS__PLUGIN_DIR . 'tpl/paymentControls.php';
		//$this->slpPaymentsCore->setIncludedButtonScript($includedButtonScript); // only include button code libs once
		$this->slpPaymentsCore->setIncludedButtonScript(true);
		wp_enqueue_script( 'badger-wallet', SlpPaymentsCore::BADGER_WALLET_JS, array(), BCHSLP_PAYMENTS_VERSION, true );
		$this->sendOrderReceivedEmail($order->get_id());
	}
	
	public function addJavascriptConfig(array $cfg): array {
		$cfg['ajaxConfirm'] = true;
		$cfg['keepTransaction'] = true;
		// total order amount must be computed manually, order object only exists after, see https://stackoverflow.com/a/46631896/5872476
		//$currency = get_woocommerce_currency();
		$cfg['totalCartValue'] = max( 0, apply_filters( 'woocommerce_calculated_total', round( WC()->cart->cart_contents_total + WC()->cart->fee_total + WC()->cart->tax_total, WC()->cart->dp ), WC()->cart ) );
		$cfg['orderID'] = 0;
		if ($this->session) {
			$cfg['orderID'] = $this->session->get("orderID", 0);
			if (is_object($cfg['orderID'])) // legacy
				$cfg['orderID'] = $cfg['orderID']->get_id();
		}
		if ($cfg['orderID'] !== 0) {
			try {
				$order = new \WC_Order($cfg['orderID']);
				if ($order->get_payment_method() === 'slp_payments_woocommerce') {
					$cfg['nonce'] = SlpPaymentsApi::createNonce('/slp-payments/v1/order-status', 'order-' . $cfg['orderID'] . '-' . $order->get_billing_email());
					$cfg['tokenID'] = $order->get_meta('tokenID');
					$cfg['address'] = $this->getOrderPaymentAddress($order);
				}
			}
			catch (\Exception $e) { // invalid order exception
			}
			$cfg['successPayImage'] = plugins_url( 'img/success-pay.gif', BCHSLP_PAYMENTS__PLUGIN_DIR . 'slp-payments.php' );
			$cfg['successPayImageFinal'] = plugins_url( 'img/success-pay-done.png', BCHSLP_PAYMENTS__PLUGIN_DIR . 'slp-payments.php' );
		}
		$cfg['checkPaymentIntervalSec'] = static::CHECK_PAYMENT_INTERVAL_SEC;
		$cfg['enabledTokens'] = $this->slpPaymentsCore->getEnabledTokens();
		$cfg['paidTxt'] = __('Paid', 'ekliptor');
		return $cfg;
	}
	
	public function addDefaultSettings(array $defaults): array {
		$addonDefaults = array(
				//'foo' => 1
		);
		return array_merge($defaults, $addonDefaults);
	}
	
	public function addWoocommerceAdminFooterCode(): void {
		$cfg = array(
				'preInstalledTokens'		=> PreInstalledTokens::getTokens(),
				'availableTokens'			=> $this->pluginSettings->get('availableTokens'),
				//'reloadSettings'			=> $this->pluginSettings->get('reloadSettings'),
				// localizations
				'tr'						=> array(
						'newToken'					=> __('New Token', 'ekliptor'),
						'confirmDeleteToken'		=> __('Do you really want to delete the selected SLP Token?', 'ekliptor'),
						'missingFields'				=> __('Please fill out the required fields Token ID and Exchange rate.', 'ekliptor'),
				),
		);
		//if ($cfg['reloadSettings'] === true)
			//$this->pluginSettings->set('reloadSettings', false);
		echo '<script type="text/javascript">var slpPaymentsAdminCfg = ' . json_encode($cfg) . ';</script>';
	}
	
	/**
	 * Check if an order has been paid. This will use (and update) the cache and is therefore preferred to a direct call of isPaidOrder().
	 * @param \WC_Order $order
	 * @param $force
	 * @return bool
	 */
	public function checkPaidOrder(\WC_Order $order, bool $force = false): bool {
		if ($order->is_paid() === true)
			return true;
		$lastCheck = $order->get_meta('lastCheckedChainPayment');
		if (!$lastCheck)
			$lastCheck = 0;
		if ($force === false && $lastCheck + static::CHECK_PAYMENT_INTERVAL_SEC > time())
			return false;
		
		$isPaid = $this->isPaidOrder($order);
		$order->add_meta_data('lastCheckedChainPayment', time(), true);
		if ($isPaid === true && $order->is_paid() !== true)
			$order->add_order_note(__("Received full payment on address: ", "ekliptor") . $this->getOrderPaymentAddress($order));
		$order->save_meta_data();
		return $isPaid;
	}
	
	//protected function isPaidOrder(int $orderID, $updateOrderErrors = true): bool {
		//if (!isset($this->session) || !isset($this->session->bchTransactionID))
			//return false;
	public function isPaidOrder(\WC_Order $order, $updateOrderErrors = false): bool {
		$created = $order->get_date_created();
		if ($created && $created->getTimestamp() + static::ORDER_UNPAID_SKIP_CHECK_SEC > time())
			return false; // order has just been created. speed up first page load
		// use the faster backend if available
		$isTokenPayment = $order->get_meta('tokenAmount') > 0.0;
		if ($this->settings['enable0Conf'] === 'yes' && $isTokenPayment === true) {
			$this->addChainTransactionID($order); // should be already done
			return $this->isPaidOrderBySlpDb($order);
		}
		
		// we use unique addresses per order. so only check the number of confirmations if 0-conf is disabled
		if (!$order->get_meta('bchTransactionID')) {
			//SlpPaymentsCore::notifyErrorExt("Can not check if order is paid without TXID", $order); // happens on call from AJAX API
			return false;
		}
		$confirmations = $this->blockchainApi->getConfirmationCount($order->get_meta('bchTransactionID'));
		if ($confirmations === -1)
			return false; // TX doesn't exist (yet)
		if ($this->settings['enable0Conf'] !== 'yes' && $confirmations < $this->pluginSettings->get('wait_confirmations'))
			return false;
		if ($isTokenPayment === true && empty($order->get_meta('tokenID')))
			return false; // not ready yet
		
		// check if the amount has been fully received
		$address = $order->get_meta('chainAddress');
		if (!$address) // should always be set
			return false;
		// check if all transactions are confirmed (usually only 1)
		$this->addChainTransactionID($order); // should be already done
		
		$transactions = $order->get_meta('addressTransactionIDs');
		$txMaxAgeSec = time() - SlpPaymentsCore::MAX_AGE_TX_ACCEPT*MINUTE_IN_SECONDS;
		if ($transactions) {
			$transactions = maybe_unserialize($transactions); // sometimes already unserialized. how?
			for ($i = 0; $i < count($transactions); $i++) { // check the first TX again to include the check for blocktime
				$confirmations = $this->blockchainApi->getConfirmationCount($transactions[$i]);
				if ($confirmations === -1)
					return false; // TX doesn't exist (yet)
				$blocktime = $this->blockchainApi->getBlocktime($transactions[$i]);
				if ($blocktime < $txMaxAgeSec)
					return false; // this TX was sent to this address before our payment request (address used by another wallet). safe to return false
				// TODO continue and this balance in check below. but we create empty addresses either way
				
				if ($this->settings['enable0Conf'] !== 'yes' && $confirmations < $this->pluginSettings->get('wait_confirmations'))
					return false;
			}
		}
		
		// all TX have enough confirmations. check balance
		if ($order->get_meta('slpPayment') === '1') {
			// $this->slpPaymentsCore->useTokenPayments() === true
			if ($isTokenPayment === true) {
				$tokenAmountReceived = $this->blockchainApi->getAddressTokenBalance($address, $order->get_meta('tokenID'));
				if ($tokenAmountReceived != $order->get_meta('tokenAmountReceived'))
					$order->add_order_note(sprintf(__("Received new (partial) payment of %s %s on address %s", "ekliptor"), $tokenAmountReceived, $order->get_meta('tokenSymbol'), $this->getOrderPaymentAddress($order)));
    			Helper::updateOrderMeta($order, 'tokenAmountReceived', $tokenAmountReceived);
				return $tokenAmountReceived >= (float)$order->get_meta('tokenAmount');
			}
		}
		$bchAmount = (float)$order->get_meta('bchAmount');
		if ($bchAmount < 0.0) {
			if ($order->get_meta('slpPayment') === '0')
				SlpPaymentsCore::notifyErrorExt("BCH payment amount is not set in order", $order);
			return false;
		}
		$maxFeeLost = min(SlpPaymentsCore::BCH_TX_FEE_TOLERANCE, $bchAmount/4.0);
		$bchAmountReceived = $this->blockchainApi->getAddressBalance($address);
		if ($bchAmountReceived != $order->get_meta('bchAmountReceived'))
			$order->add_order_note(sprintf(__("Received new (partial) payment of %s BCH on address %s", "ekliptor"), $bchAmountReceived, $this->getOrderPaymentAddress($order)));
		Helper::updateOrderMeta($order, 'bchAmountReceived', $bchAmountReceived);
		if (SlpPaymentsCore::toSatoshis($bchAmount) > SlpPaymentsCore::BCH_DUST_LIMIT_SAT && $bchAmountReceived + $maxFeeLost < $bchAmount)
			return false;
		$this->clearCustomSessionVariables();
		// TODO display errors as HTML if we call this via HTTP from client. currently we just wait (for confirmations) and let the user reload manually
		return true;
	}
	
	protected function isPaidOrderBySlpDb(\WC_Order $order): bool {
		if ($this->slpDbApi === null)
			$this->slpDbApi = $this->slpPaymentsCore->createBlockchainApiInstance('SlpDbApi', $this->pluginSettings->get('slpDbApiUrl'));
		$transactions = $order->get_meta('addressTransactionIDs');
		$txMaxAgeSec = time() - SlpPaymentsCore::MAX_AGE_TX_ACCEPT*MINUTE_IN_SECONDS;
		if ($transactions) {
			$transactions = maybe_unserialize($transactions); // sometimes already unserialized. how?
			for ($i = 0; $i < count($transactions); $i++) { // check the first TX again to include the check for blocktime
				$confirmations = $this->slpDbApi->getConfirmationCount($transactions[$i]);
				if ($confirmations === -1)
					return false; // TX doesn't exist (yet)
				$blocktime = $this->blockchainApi->getBlocktime($transactions[$i]);
				if ($blocktime < $txMaxAgeSec)
					return false; // this TX was sent to this address before our payment request (address used by another wallet). safe to return false
			}
		}
		
		$slpAddress = $order->get_meta('slpAddress');
		$tokenAmountReceived = $this->slpDbApi->getAddressTokenBalance($slpAddress, $order->get_meta('tokenID'));
    	Helper::updateOrderMeta($order, 'tokenAmountReceived', $tokenAmountReceived);
    	if ($tokenAmountReceived >= (float)$order->get_meta('tokenAmount')) {
			$this->clearCustomSessionVariables();
			return true;
    	}
		return false;
	}
	
	public function getSuccessPaymentStatus(): string {
		return 'processing'; // TODO add option to set it to completed immediately (for digial goods)? but processing is default
	}
	
	public function checkTransactionConfirmations() {
		$this->pluginSettings->set('lastCheckedTransactions', time());
		$orders = $this->getPendingOrders();
		foreach ($orders as $order) {
			$this->addChainTransactionID($order);
			if ($this->checkPaidOrder($order) === false)
				continue;
			$order->update_status($this->getSuccessPaymentStatus(), '');
			//$order->save_meta_data(); // not needed when calling update_status()
		}
	}
	
	/**
	 * This function adds our BCH and SLP data to the invoice (before payment) and the payment confirmation (after payment).
	 * @param \WC_Order $order
	 * @param bool $sent_to_admin
	 * @param bool $plain_text
	 * @param \WC_Email $email
	 */
	public function addBlockchainDataToEmail(\WC_Order $order, bool $sent_to_admin, bool $plain_text, \WC_Email $email) {
		if ($order->get_payment_method() !== 'slp_payments_woocommerce')
			return; // the user chose another payment method
		//if ($order->is_paid() !== true) // just to be sure
			//return; // also called for customer invoice
		if ($order->is_paid() === true && !$order->get_meta('bchTransactionID')) {
			//SlpPaymentsCore::notifyErrorExt("TXID not available when sending confirmation order email", $order);
			return;
		}
		
		//$useTokenPayments = $this->slpPaymentsCore->useTokenPayments() === true; // setting might change til payment
		$useTokenPayments = $order->get_meta('tokenAmount') > 0.0;
		$totalAmount = $useTokenPayments === true ? (float)$order->get_meta('tokenAmount') : (float)$order->get_meta('bchAmount');
		if ($totalAmount == 0.0)
			return; // shouldn't happen
		$decimals = $this->pluginSettings->get('display_digits_bch');
		if ($useTokenPayments === true) {
			$slpToken = $this->slpPaymentsCore->getTokenBySymbol($order->get_meta('tokenSymbol'));
			$decimals = $slpToken['decimals'];
		}
		$tplVars = array(
				'paid' => $order->is_paid() === true,
				'txid' => $order->get_meta('bchTransactionID'),
				'amountDisplay' => number_format($totalAmount, $decimals),
				'tickerDisplay' => $useTokenPayments === true ? $order->get_meta('tokenSymbol') : 'BCH',
				'recAddress' => $useTokenPayments === true ? $order->get_meta('slpAddress') : $order->get_meta('chainAddress')
		);
		ob_start();
		include BCHSLP_PAYMENTS__PLUGIN_DIR . 'tpl/email/invoice.php'; // TODO add qr code, badger button,... ?
		$docHtml = ob_get_contents();
		ob_end_clean();
		echo $docHtml;
	}
	
	public function sendOrderReceivedEmail(int $orderID) {
		if ($this->settings['sendPaymentInstructions'] !== 'yes')
			return;
		$order = new \WC_Order($orderID);
		if ($order->get_payment_method() !== 'slp_payments_woocommerce')
			return; // the user chose another payment method
		if ($order->get_meta('sentChainOrderMail') === '1')
			return; // don't send email twice (if admin changes order status back and forth)
		if ($order->is_paid() === true)
			return; // already paid, WP is late, avoid confusion
		
		$mailer = WC()->mailer();
		/* // use the existing WC invoice template
		$recipient = $order->get_billing_email();
		$subject = __('Payment instructions for your order', 'ekliptor');
		$content = $this->getOrderReceivedMailContent($order, $subject, $mailer);
		$headers = "Content-Type: text/html\r\n";
		$mailer->send( $recipient, $subject, $content, $headers );
		*/
		$mailer->customer_invoice($order);
		
		$order->add_meta_data('sentChainOrderMail', '1', true);
		$order->add_order_note(__("Sent payment instructions email to " . $order->get_billing_email(), "ekliptor"));
		$order->save_meta_data();
	}
	
	/*
	protected function getOrderReceivedMailContent(\WC_Order $order, string $heading, \WC_Emails $mailer): string {
		//$template = 'emails/customer-processing-order.php';
		$template = 'emails/customer-invoice.php';
	    return wc_get_template_html( $template, array(
	        'order'         => $order,
	        'email_heading' => $heading,
	        'sent_to_admin' => false,
	        'plain_text'    => false,
	        'email'         => $mailer
	    ) );
	}
	*/
	
	public function addChainTransactionID(\WC_Order $order) {
		// for all orders not paid with BadgerWallet we don't have TX ID
		// so we must watch the associated 1-time address
		$bchTxID = $order->get_meta('bchTransactionID');
		if (!empty($bchTxID))
			return; // already set
		else if (empty($order->get_meta('tokenID')))
			return; // not ready yet
		$address = $order->get_meta('chainAddress');
		if (empty($address)) {
			//SlpPaymentsCore::notifyErrorExt("Found WC order without chain address", $order);
			return; // first call
		}
		// an address might have multiple TXIDs, although generally our 1-time addresses will only have 1
		$addressDetails = array();
		if ($order->get_meta('slpPayment') === '1')
			$addressDetails = $this->blockchainApi->getSlpAddressDetails($order->get_meta('slpAddress'), $order->get_meta('tokenID'));
		else
			$addressDetails = $this->blockchainApi->getAddressDetails($address);
		if ($addressDetails === null || empty($addressDetails->transactions))
			return; // no TX yet
		$order->add_meta_data('bchTransactionID', $addressDetails->transactions[0], true); // the first is usucally the only one
		$order->add_meta_data('addressTransactionIDs', serialize($addressDetails->transactions), true);
		$order->save_meta_data();
	}
	
	public function generateQrCodeForOrder(\WC_Order $order): string {
		$isSlpPayment = $order->get_meta('slpPayment') === '1';
		$slpToken = $isSlpPayment === true ? $this->slpPaymentsCore->getTokenBySymbol($order->get_meta('tokenSymbol')) : null;
		$address = $isSlpPayment === false ? $order->get_meta('chainAddress') : $order->get_meta('slpAddress');
		$bchRemaining = max(0.0, (float)$order->get_meta('bchAmount') - (float)$order->get_meta('bchAmountReceived'));
		$tokenRemaining = max(0.0, (float)$order->get_meta('tokenAmount') - (float)$order->get_meta('tokenAmountReceived'));
		return $this->slpPaymentsCore->generateQrCodeForAddress($order->get_meta('slp-txid'), $address, $order->get_total(), $bchRemaining,
				$tokenRemaining, $order->get_meta('tokenID'), $slpToken ? $slpToken['decimals'] : $this->pluginSettings->get('display_digits_bch'));
	}
	
	public function generatePaymentUriForOrder(\WC_Order $order): string {
		$isSlpPayment = $order->get_meta('slpPayment') === '1';
		$slpToken = $isSlpPayment === true ? $this->slpPaymentsCore->getTokenBySymbol($order->get_meta('tokenSymbol')) : null;
		$address = $isSlpPayment === false ? $order->get_meta('chainAddress') : $order->get_meta('slpAddress');
		$bchRemaining = max(0.0, (float)$order->get_meta('bchAmount') - (float)$order->get_meta('bchAmountReceived'));
		$tokenRemaining = max(0.0, (float)$order->get_meta('tokenAmount') - (float)$order->get_meta('tokenAmountReceived'));
		if ($slpToken)
    		return $this->slpPaymentsCore->createPaymentURI($address, $bchRemaining, $tokenRemaining, $slpToken['id'], isset($slpToken['decimals']) ? $slpToken['decimals'] : 8);
    	return $this->slpPaymentsCore->createPaymentURI($address, $bchRemaining, 0.0);
	}
	
	public function getOrderPaymentAddress(\WC_Order $order): string {
		$useTokenPayments = $order->get_meta('tokenAmount') > 0.0; // or use $order->get_meta('slpPayment') === '1'
		$address = $useTokenPayments === true ? $order->get_meta('slpAddress') : $order->get_meta('chainAddress');
		return empty($address) ? "" : $address;
	}
	
	public function generate_plugin_button_html(string $key, $data ) {
		// this method MUST be named exactly like this because it's called from WC parent class via generate_{type}_html
		return $this->tpl->generateWoocommerceButtonHtml($this, $key, $data);
	}
	
	protected function getPendingOrders(): array {
		// orders will be cancelled by Woocommerce after 1 hour. hook woocommerce_cancel_unpaid_order filter to change that
		// https://docs.woocommerce.com/wc-apidocs/source-function-wc_cancel_unpaid_orders.html#904
		// TODO add a setting? already present in Woocommerce core as setting
		$args = array(
				'status' => 'pending',
				'payment_method' => 'slp_payments_woocommerce',
				//'created_via' => '',
				'orderby' => 'modified',
    			'order' => 'DESC',
				'limit' => 5000, // TODO more within 1h needed?
		);
		$orders = wc_get_orders( $args );
		return $orders;
	}
	
	protected function clearCustomSessionVariables() {
		// shouldn't be needed since the whole session gets destroyed eventually, but let's be clean
		if (!$this->session)
			return;
		$keys = array(/*"orderID", */"bchAddress"); // don't remove the order ID because we need it to show success message
		foreach ($keys as $key) {
			//$this->session->__unset($key); // unreliable // TODO why?
			$this->session->set($key, null);
		}
	}
	
	protected function getFormFieldArray(): array {
		$fields = array(
				'enabled' => array(
						'title' 		=> __('Enable this SLP payments plugin', 'ekliptor'),
						'type'			=> 'checkbox',
						'description'	=> '',
						'default'		=> 'yes'
				),
				'xPub' => array(
						'title' 		=> __('xPub', 'ekliptor'),
						'type'			=> 'text',
						'description'	=> __('The extended public key (xPub) to create addresses from. In Electron Cash this is called "Master Public Key" and can be found under Wallet -> Information.', 'ekliptor'),
						'default'		=> '',
						'desc_tip'		=> true
				),
				'enable0Conf' => array(
						'title' 		=> __('Enable 0-conf payments', 'ekliptor'),
						'type'			=> 'checkbox',
						'description'	=> sprintf(__('If enabled, Woocommerce orders will be instantly marked as paid once the transaction has been broadcast to the network.<br>Otherwise transactions will be marked as pending until %s confirmations have been received.', 'ekliptor'), $this->pluginSettings->get('wait_confirmations')),
						'default'		=> 'yes',
						'desc_tip'		=> true
				),
				'title' => array(
						'title' 		=> __('Payment Method Title', 'ekliptor'),
						'type'			=> 'text',
						'description'	=> __('The payment method title your customers will see in your shop.', 'ekliptor'),
						'default'		=> $this->getFrontendDefaultTitle(),
						'desc_tip'		=> true
				),
				'description' => array(
						'title' 		=> __('Description', 'ekliptor'),
						'type'			=> 'text',
						'description'	=> __('The payment method description your customers will see in your shop.', 'ekliptor'),
						'default'		=> $this->getFrontendDefaultDescription(),
						'desc_tip'		=> true
				),
				'enableBadger' => array(
						'title' 		=> __('Enable BadgerWallet', 'ekliptor'),
						'type'			=> 'checkbox',
						'description'	=> __('Show a BadgerButton for the BadgerWallet browser extension as payment option.', 'ekliptor'),
						'default'		=> 'yes',
						'desc_tip'		=> true
				),
				'blockchainApi' => array(
						'title' 		=> __('BCH Blockchain API', 'ekliptor'),
						'type'			=> 'text',
						'description'	=> __('The REST backend to connect to for the BCH blockchain data. You may run your own backend: https://github.com/Bitcoin-com/rest.bitcoin.com', 'ekliptor'),
						'default'		=> $this->pluginSettings->get('blockchain_rest_url'),
						'desc_tip'		=> true
				),
				'slpDbApi' => array(
						'title' 		=> __('SLPDB API', 'ekliptor'),
						'type'			=> 'text',
						'description'	=> __('SLPDB API REST backend to connect to for SLP blockchain data. You may run your own backend: https://github.com/simpleledger/SLPDB', 'ekliptor'),
						'default'		=> $this->pluginSettings->get('slpDbApiUrl'),
						'desc_tip'		=> true
				),
				'slpSocketAddress' => array(
						'title' 		=> __('SLP Socket API', 'ekliptor'),
						'type'			=> 'text',
						'description'	=> __('SLP Socket API backend to connect to for SLP blockchain data. You may run your own backend: https://github.com/fountainhead-cash/slpsockserve', 'ekliptor'),
						'default'		=> $this->pluginSettings->get('slpSocketAddress'),
						'desc_tip'		=> true
				),
				'sendPaymentInstructions' => array(
						'title' 		=> __('Send payment instructions email', 'ekliptor'),
						'type'			=> 'checkbox',
						'description'	=> __('If enabled, this plugin will send Woocommerce invoice templates (see Woocommerce -> Settings -> Emails) with payment instructions as email.', 'ekliptor'),
						'default'		=> 'yes',
						'desc_tip'		=> true
				),
				
				// Token config
				'allowBchPayments' => array(
					'title' 		=> __('Also allow BCH payments', 'ekliptor'),
					'type'			=> 'checkbox',
					'description'	=> 'Enable this to let the user choose between BCH and SLP payments. Disable this to only allow SLP payments.',
					'default'		=> 'yes',
					'desc_tip'		=> true
				),
				'tokenSelect' => array(
						'title' 		=> __('SLP Token', 'ekliptor'),
						'type'			=> 'select',
						'description'	=> __('Select the SLP token you want to configure below.', 'ekliptor'),
						'desc_tip'		=> true,
						//'custom_attributes' => array('id' => 'slp-token-select'), // doesn't show up
						//'class'			=> 'slp-token-select', // works, but use official id naming scheme: woocommerce_slp_payments_woocommerce_tokenSelect
						'options'		=> $this->getTokenSelectOptions()
				),
				/* // moved to main Settings class
				'availableTokens' => array(
						'title' 		=> '',
						'type'			=> 'text',
						'description'	=> '',
						'disabled'		=> true,
						'default'		=> array_keys($this->getTokenSelectOptions()),
						'desc_tip'		=> false
				),
				*/
				// hidden form fields for add/delete buttons to submit them
				'deleteCurrentToken' => array(
						'title' 		=> '',
						'type'			=> 'text',
						'description'	=> '',
						'disabled'		=> true,
						'default'		=> '',
						'desc_tip'		=> false
				),
				'addingNewToken' => array(
						'title' 		=> '',
						'type'			=> 'text',
						'description'	=> '',
						'disabled'		=> true,
						'default'		=> '',
						'desc_tip'		=> false
				),
				'allTokenJson' => array(
						'title' 		=> '',
						'type'			=> 'text',
						'description'	=> '',
						'disabled'		=> true,
						'default'		=> '',
						'desc_tip'		=> false
				),
				
				'addToken' => array(
						'title' 		=> __('Add Token', 'ekliptor'),
						'button_txt'	=> __('Add new Token', 'ekliptor'),
						'type'			=> 'plugin_button',
						'id'			=> 'slp-add-token',
						'description'	=> __('Click this to add a new SLP token as a payment option.', 'ekliptor'),
						'desc_tip'		=> true
				),
				'deleteToken' => array(
						'title' 		=> __('Delete Token', 'ekliptor'),
						'button_txt'	=> __('Delete current Token', 'ekliptor'),
						'type'			=> 'plugin_button',
						'id'			=> 'slp-delete-token',
						'description'	=> __('Click this to delete the currently selected SLP token from your payment options.', 'ekliptor'),
						'desc_tip'		=> true
				),
				// 'enabledToken' => array(
				// 		'title' 		=> __('Enable SLP token payments', 'ekliptor'),
				// 		'type'			=> 'checkbox',
				// 		'description'	=> 'Enable payments in a specified SLP token to give customers a discount by splitting the payment.',
				// 		'default'		=> 'no',
				// 		'desc_tip'		=> true
				// ),
				'enableCurrentToken' => array(
					'title' 		=> __('Enable this Token', 'ekliptor'),
					'type'			=> 'checkbox',
					'description'	=> 'Enable this to let users pay in the currently selected token.',
					'default'		=> 'yes',
					'desc_tip'		=> true,
				),
				'tokenID' => array( // TODO make this and ticker symbol disabled if a value exists? editing it means adding a new token
						'title' 		=> __('Token ID', 'ekliptor'),
						'type'			=> 'text',
						'description'	=> __('The ID uniquely identifying the token (as hex string).', 'ekliptor'),
						'default'		=> '',
						'desc_tip'		=> true
				),
				'tokenRateUSD' => array( // TODO other currencies
						'title' 		=> __('Exchange rate (USD)', 'ekliptor'),
						'type'			=> 'text',
						'description'	=> __('The rate (price) of 1 token in USD.', 'ekliptor'),
						'default'		=> 1.0,
						'desc_tip'		=> true
				),
				'tokenTickerSymbol' => array(
						'title' 		=> __('Ticker Symbol', 'ekliptor'),
						'type'			=> 'text',
						'description'	=> __('The ticker symbol of the token your shop accepts payments in. (Optional, defaults to the on-chain value.)', 'ekliptor'),
						'default'		=> '',
						'desc_tip'		=> true
				),
				'tokenName' => array(
						'title' 		=> __('Token Name', 'ekliptor'),
						'type'			=> 'text',
						'description'	=> __('The name of the token your shop accepts payments in. (Optional, defaults to the on-chain value.)', 'ekliptor'),
						'default'		=> '',
						'desc_tip'		=> true
				),
				'tokenImage' => array(
						'title' 		=> __('Token icon', 'ekliptor'),
						'type'			=> 'text',
						'description'	=> __('The image/icon URL of the token.', 'ekliptor'),
						'default'		=> '',
						'desc_tip'		=> true
				),
				'tokenWebsite' => array(
						'title' 		=> __('Website', 'ekliptor'),
						'type'			=> 'text',
						'description'	=> __('The link to the website of the token for more information.', 'ekliptor'),
						'default'		=> '',
						'desc_tip'		=> true
				),
		);
		if ($this->showDebugInfo === true) {
			unset($this->settings['_dbgLastAddress']); // read-only
			$fields['_dbgLastAddress'] = array(
					'title' 		=> __('Last address', 'ekliptor'),
					'type'			=> 'text',
					'description'	=> __('The last address generated from the supplied xPub.', 'ekliptor'),
					'default'		=> $this->pluginSettings->get('lastAddress'),
					'desc_tip'		=> true,
					'class'			=> 'slp-debug-info'
				);
			unset($this->settings['_dbgLastHdPath']); // read-only
			$fields['_dbgLastHdPath'] = array(
					'title' 		=> __('Last HDPath', 'ekliptor'),
					'type'			=> 'text',
					'description'	=> __('The HDPath with which to generate the last address.', 'ekliptor'),
					'default'		=> sprintf($this->pluginSettings->get('hdPathFormat'), $this->pluginSettings->get('addressCount')),
					'desc_tip'		=> true,
					'class'			=> 'slp-debug-info'
				);
			unset($this->settings['_dbgLastBchRateUpdate']); // read-only
			$lastUpdate = $this->pluginSettings->get('last_rate_update');
			$lastUpdateText = $this->tpl ? sprintf("%s %s", $this->tpl->getDate($lastUpdate, false), $this->tpl->getTime($lastUpdate, false)) : '';
			$fields['_dbgLastBchRateUpdate'] = array(
					'title' 		=> __('Last BCH rate update', 'ekliptor'),
					'type'			=> 'text',
					'description'	=> __('The time when the BCH rate was last updated by the WordPress cron.', 'ekliptor'),
					'default'		=> $lastUpdateText,
					'desc_tip'		=> true,
					'class'			=> 'slp-debug-info'
				);
		}
		return $fields;
	}
	
	protected function getTokenSelectOptions() {
		$availableTokens = $this->pluginSettings->get('availableTokens');
		$options = array();
		foreach ($availableTokens as $symbol => $token) {
			if (isset($token['name'])) // filter invalid tokens
				$options[$symbol] = $token['name'];
		}
		return $options;
	}
		
	protected function getFrontendDefaultTitle(): string {
		return __('SLP Tokens & Bitcoin Cash Payments', 'ekliptor');
	}
	
	protected function getFrontendDefaultDescription(): string {
		return __('Instant and low fee Bitcoin Cash and SLP token payment on-chain without going through a 3rd party.', 'ekliptor');
	}
}
?>