<?php
/*
 * Plugin Name: SLP Payments for WooCommerce
 * Plugin URI: https://simpleledger.io/
 * Description: Accept SLP tokens and Bitcoin Cash in your online store as payment.
 * Version: 1.2.26
 * Author: Simple Ledger Inc
 * Author URI: https://twitter.com/SimpleLedger
 * License: GPLv3
 * Text Domain: ekliptor
 * 
 * WC requires at least: 3.0
 * WC tested up to: 3.6
 */

use Ekliptor\Wordpress\DatabaseMigration;
use Ekliptor\Wordpress\SlpPayments;
use Ekliptor\Wordpress\SlpPaymentsAdmin;
use Ekliptor\Wordpress\SlpPaymentsApi;

// Make sure we don't expose any info if called directly
if (! defined( 'ABSPATH' )) {
	echo 'Hi there!  I\'m just a plugin, not much I can do when called directly.';
	exit ();
}

define ( 'BCHSLP_PAYMENTS_VERSION', '1.2.26' );
define ( 'BCHSLP_PAYMENTS__MINIMUM_WP_VERSION', '4.7' );
define ( 'BCHSLP_PAYMENTS__PLUGIN_DIR', plugin_dir_path ( __FILE__ ) );

function bchslp_woocommerce_load() {
	if (class_exists(/*'WooCommerce'*/'WC_Payment_Gateway') === false) {
		add_action('admin_notices', function() { // TODO check for Woocommerce version?
			load_plugin_textdomain ( 'ekliptor' );
			$message = '<strong>' . esc_html__ ( 'For SLP Payments to work you must have Woocommerce installed and activated in your WordPress site.', 'ekliptor' ) . '</strong><br> ' . sprintf(esc_html__ ( 'Please download and activate the Woocommerce plugin from your WordPress admin panel in the plugin installer or from %s and then install activate this plugin again.', 'ekliptor'),
				sprintf('<a target="_blank" href="%s">%s</a>', 'https://woocommerce.com/', __('here', 'ekliptor')) );
			echo sprintf('<div class="notice notice-%s%s">
	          %s
	         </div>', 'error', '', $message);
		});
		return;
	}
	
	if (is_admin ()/* || (defined ( 'WP_CLI' ) && WP_CLI)*/) {
		require_once (BCHSLP_PAYMENTS__PLUGIN_DIR . 'classes/SlpPaymentsAdmin.php'); // PHP functions + classes always have the global scope
	}
	/*
	SlpPayments::check_plugin_activation();
	SlpPaymentsAdmin::getInstance(SlpPayments::getInstance());
	add_action ( 'init', array (
			SlpPayments::getInstance(),
			'init' 
	), 11 );
	*/
	
	SlpPaymentsApi::getInstance(/*SlpPayments::getInstance()*/);
	add_action ( 'rest_api_init', array (
			SlpPaymentsApi::getInstance(),
			'init' 
	) );
	
	/*
	if (is_admin ()/* || (defined ( 'WP_CLI' ) && WP_CLI)***) {
		SlpPaymentsAdmin::getInstance(SlpPayments::getInstance());
		add_action ( 'init', array (
				SlpPaymentsAdmin::getInstance(),
				'init' 
		), 11 );
	}
	*/
	add_filter('woocommerce_payment_gateways', array('Ekliptor\Wordpress\SlpPayments', 'addWoocommerceGateway'), 10, 1);
}

/* // sub plugin: we must check for activation on every startup after plugins are loaded with autoload option
register_activation_hook ( __FILE__, array (
		'Ekliptor\Wordpress\SlpPayments',
		'plugin_activation' 
) );*/
register_deactivation_hook ( __FILE__, array (
		'Ekliptor\Wordpress\SlpPayments',
		'plugin_deactivation' 
) );

add_action( 'plugins_loaded', function () {
	require_once (BCHSLP_PAYMENTS__PLUGIN_DIR . 'data.php');
	require_once (BCHSLP_PAYMENTS__PLUGIN_DIR . 'functions.php');
	require_once (BCHSLP_PAYMENTS__PLUGIN_DIR . 'classes/SlpPayments.php');
	require_once (BCHSLP_PAYMENTS__PLUGIN_DIR . 'classes/SlpPaymentsApi.php');
	require_once (BCHSLP_PAYMENTS__PLUGIN_DIR . 'api.php');
	bchslp_woocommerce_load();
}, 100 );

require_once (BCHSLP_PAYMENTS__PLUGIN_DIR . 'classes/DatabaseMigration.php');
update_option ( 'slp_payments_installed', '1.2.22' );
DatabaseMigration::checkAndMigrate();

// crons
add_action( 'init', function() {
	if (defined('DOING_CRON') && class_exists('WC_Payment_Gateway') === true)
		SlpPayments::setupCronHooks(); // TODO this gets called before plugins_loaded, so we might have 2 instances for crons
});

// privacy policy // TODO move to admin page if we add a top-level admin page
add_action( 'admin_init', function() {
	if ( ! function_exists( 'wp_add_privacy_policy_content' ) )
    	return;
    $content = sprintf(
        	__( 'This plugin loads code API code from bitcoin.com to interact with BadgerWallet - a browser extension for BCH payments. 
				This does not require an account or signup at their service. Their privacy policy is available here: https://www.bitcoin.com/privacy-policy/',
        			'ekliptor' )
    );
    wp_add_privacy_policy_content('SLP Payments', wp_kses_post( wpautop( $content, false ) ) );
});
?>