
export class SlpDb {
    constructor() {

    }

    public getAddressBalanceQuery(address: string, tokenID: string): string {
        return `{
          "v": 3,
          "q": {
            "find": {"slp.detail.tokenIdHex": "${tokenID}", "slp.detail.outputs.address": "${address}"}
          }
        }`;
    }

    public getTotalOutputAmount(data: any, toAddress: string, tokenID: string): number {
        /** // TODO add/import interface
         * data: {
            "type": "block",
            "data": [{
                "txid": "02d6189c275d9193bf30b15a72980390519dadebf625415c56a0ab168a5e7994",
                "slp": {
                    "valid": true,
                    "detail": {
                        "decimals": 4,
                        "tokenIdHex": "02d6189c275d9193bf30b15a72980390519dadebf625415c56a0ab168a5e7994",
                        "transactionType": "GENESIS",
                        "versionType": 1,
                        "documentUri": "https://bitcoin.tk",
                        "documentSha256Hex": null,
                        "symbol": "BTK",
                        "name": "BitcoinTK",
                        "txnBatonVout": null,
                        "txnContainsBaton": false,
                        "outputs": [{
                            "address": "simpleledger:qqw7qrvak6nu2y7k39c0lt9wyv9wwkuh7ygj4e75ul",
                            "amount": "21000000"
                        }]
                    },
                    "invalidReason": null,
                    "schema_version": 73
                }
            },
         */
        let totalOutAmount = 0.0;
        if (!data || Array.isArray(data.data) === false)
            return totalOutAmount; // invalid?
        data.data.forEach((tx) => {
            if (!tx.slp || tx.slp.valid !== true || !tx.slp.detail)
                return;
            const txDetails = tx.slp.detail;
            if (txDetails.tokenIdHex !== tokenID)
                return; // for another token
            if (Array.isArray(txDetails.outputs) === false)
                return; // invalid, shouldn't happen
            txDetails.outputs.forEach((out) => {
                if (out.address !== toAddress)
                    return // to another receiver
                const amount = parseFloat(out.amount);
                if (amount !== Number.NaN && amount > 0.0)
                    totalOutAmount += amount;
            });
        });
        return totalOutAmount;
    }

    // ################################################################
    // ###################### PRIVATE FUNCTIONS #######################
}
